@@ -380,7 +380,7 @@ class AutoPdb:
         if exc_val is None:
             return
         if isinstance(exc_val, bdb.BdbQuit):
-            print("Post mortem is skipped because the exception is from Pdb.")
+            logger.info("Post mortem is skipped because the exception is from Pdb.")
         else:
             traceback.print_exc()
             pdb.post_mortem(exc_tb)
@@ -594,6 +594,12 @@ def pt_profile(func):
         return func
 
 
+def readable_size(num_bytes):
+    from math import isnan
+    from calmsize import size as calmsize
+    return '' if isnan(num_bytes) else '{:.2f}'.format(calmsize(num_bytes))
+
+
 # Monkey Patch pytorch_memlab:
 if 'MEM_PROFILE' in os.environ.keys():
     mem_profile_spec = os.environ['MEM_PROFILE']
@@ -605,7 +611,7 @@ if 'MEM_PROFILE' in os.environ.keys():
 
     if mem_profile_spec == 1:
 
-        from pytorch_memlab.line_profiler.line_records import RecordsDisplay, readable_size
+        from pytorch_memlab.line_profiler.line_records import RecordsDisplay
 
         def new_repr(self):
             recorded_func_names = self._line_records.index.levels[0]
@@ -687,13 +693,22 @@ def mem_profile_class():
 
 def memory_usage(tensor: "torch.Tensor" = None):
     import torch
+
     if isinstance(tensor, torch.Tensor):
-        size_mb = tensor.element_size() * tensor.nelement() / 1024 / 1024
-        size_storage = sys.getsizeof(tensor.storage()) / 1024 / 1024
-        return f"Torch tensor {list(tensor.size())}, logical {size_mb:.2f}MB, actual {size_storage:.2f}MB."
+        size_mb = readable_size(tensor.element_size() * tensor.nelement())
+        size_storage = readable_size(sys.getsizeof(tensor.storage()))
+        return f"Torch tensor {list(tensor.size())}, logical {size_mb}, actual {size_storage}."
     elif tensor is None:
-        from pytorch_memlab.line_profiler.line_records import readable_size
-        return f"[Active {readable_size(torch.cuda.memory_allocated())}]"
+        pytorch_active = readable_size(torch.cuda.memory_allocated())
+        try:
+            smi_active = readable_size(get_gpu_status("localhost")[0].gpu_mem_byte)
+        except Exception:
+            smi_active = "-"
+        return f"[Active {pytorch_active}, SMI = {smi_active}]"
+    elif isinstance(tensor, str) and tensor == "lost":
+        pytorch_reserved = torch.cuda.memory_reserved()
+        smi_active = get_gpu_status("localhost")[0].gpu_mem_byte
+        return f"[Lost {readable_size(smi_active - pytorch_reserved)}]"
     else:
         return "Memory usage not supported."
 
@@ -714,6 +729,8 @@ class ComputeDevice:
     def __init__(self):
         self.server_name = None
         self.gpu_id = None
+        self.gpu_mem_byte = None
+        self.gpu_mem_total = None
         self.gpu_mem_usage = None
         self.gpu_compute_usage = None
         self.processes = []
@@ -752,11 +769,15 @@ def get_gpu_status(server_name, get_process_info: bool = False, use_nvml: bool =
             except pynvml.NVMLError_GpuIsLost:
                 print(f"Warning: GPU {gpu_id} is lost.")
                 continue
+            handle_rate = pynvml.nvmlDeviceGetUtilizationRates(handle)
+            handle_memory = pynvml.nvmlDeviceGetMemoryInfo(handle)
             cur_dev = ComputeDevice()
             cur_dev.server_name = server_name
             cur_dev.gpu_id = gpu_id
-            cur_dev.gpu_compute_usage = pynvml.nvmlDeviceGetUtilizationRates(handle).gpu / 100
-            cur_dev.gpu_mem_usage = pynvml.nvmlDeviceGetUtilizationRates(handle).memory / 100
+            cur_dev.gpu_compute_usage = handle_rate.gpu / 100
+            cur_dev.gpu_mem_usage = handle_rate.memory / 100
+            cur_dev.gpu_mem_byte = handle_memory.used
+            cur_dev.gpu_mem_total = handle_memory.total
             all_devs.append(cur_dev)
         pynvml.nvmlShutdown()
         proc_info = []
@@ -769,6 +790,8 @@ def get_gpu_status(server_name, get_process_info: bool = False, use_nvml: bool =
             cur_dev.server_name = server_name
             cur_dev.gpu_id = gpu_id
             cur_dev.gpu_mem_usage = int(cur_mem) / int(all_mem)
+            cur_dev.gpu_mem_byte = cur_mem * 1024 * 1024
+            cur_dev.gpu_mem_total = all_mem * 1024 * 1024
             cur_dev.gpu_compute_usage = int(cur_util) / 100
             all_devs.append(cur_dev)
 
